#include "charpool.h"

static char *charpool[16];
static int currentcharpool;
static int currentcharpoolsz;
static char *nextchar;
static int charpool_initialized = 0;

static int cp_init(void) {
  /* Create our char pool */
  currentcharpool = 0;
  currentcharpoolsz = 16384;
  nextchar = charpool[0] = (char *) safe_malloc(currentcharpoolsz);
  charpool_initialized = 1;
  return 0;
}

char *cp_alloc(int sz) {
  char *p;

  if (!charpool_initialized) cp_init();

  if ((nextchar - charpool[currentcharpool]) + sz <= currentcharpoolsz) {
    p = nextchar;
    nextchar += sz;
    return p;
  }
  /* Doh!  We've got to make room */
  if (++currentcharpool > 15) {
    fatal("Character Pool is out of buckets!");
  }
  currentcharpoolsz <<= 1;

  nextchar = charpool[currentcharpool] = (char *)
    safe_malloc(currentcharpoolsz);

 return cp_alloc(sz);
 
}

char *cp_strdup(const char *src) {
const char *p = src;
char *q = nextchar;
/* end points to the first illegal char */
char *end = charpool[currentcharpool] + currentcharpoolsz;

 while((nextchar < end) && *p) {
   *nextchar++ = *p++;
 }

 if (nextchar < end) {
   /* Goody, we have space */
   *nextchar++ = '\0';
   return q;
 }

 /* Doh!  We ran out -- need to allocate more */
 currentcharpool++; /* Will never get too high */
 currentcharpoolsz <<= 2;

 nextchar = charpool[currentcharpool] = (char *)
   safe_malloc(currentcharpoolsz);
 
 return cp_strdup(src);
}
